package com.raffaeleconforti.log.util;

import org.deckfour.xes.factory.XFactoryNaiveImpl;
import org.deckfour.xes.model.*;
import org.deckfour.xes.factory.XFactory;

import java.util.HashMap;
import java.util.Map;

/**
 * Created by Raffaele Conforti (conforti.raffaele@gmail.com) on 16/02/15.
 */
public class LogCloner {

    private final Map<String, String> cachedXAttributeName = new HashMap<String, String>();
    private final Map<String, HashMap<String, String>> cachedLiteralAttribute = new HashMap<String, HashMap<String, String>>();
//    private final Map<String, HashMap<Date, Date>> cachedTimestampAttribute = new HashMap<>();
    private final Map<String, HashMap<Double, Double>> cachedContinuousAttribute = new HashMap<String, HashMap<Double, Double>>();
    private final Map<String, HashMap<Long, Long>> cachedDiscreteAttribute = new HashMap<String, HashMap<Long, Long>>();
    private final XFactory factory;

    public LogCloner() {
//        factory = new XFactoryMemoryImpl();
        factory = new XFactoryNaiveImpl();
    }

    public LogCloner(XFactory factory) {
        this.factory = factory;
    }

    public XLog cloneLog(XLog log) {
        XLog newLog = factory.createLog(log.getAttributes());
        for (XTrace trace : log) {
            XTrace newTrace = getXTrace(trace);
            newLog.add(newTrace);
        }

        return newLog;
    }

    public String getXAttributeName(String attributeName) {
        String attName = attributeName;
        if ((attName = cachedXAttributeName.get(attName)) == null) {
            attName = attributeName;
            cachedXAttributeName.put(attName, attName);
        }
        return attName;
    }

    public XAttribute getXAttribute(String attributeName, XAttribute attribute) {

        String attName = getXAttributeName(attributeName);

        XAttribute att;
        if (attribute instanceof XAttributeLiteral) {
            String attValue = ((XAttributeLiteral) attribute).getValue();
            HashMap<String, String> map;
            if ((map = cachedLiteralAttribute.get(attName)) == null) {
                map = new HashMap<String, String>();
                cachedLiteralAttribute.put(attName, map);
            }
            if ((attValue = map.get(attValue)) == null) {
                attValue = ((XAttributeLiteral) attribute).getValue();
                map.put(attValue, attValue);
            }
            att = factory.createAttributeLiteral(attName, attValue, null);
//        } else if (attribute instanceof XAttributeTimestamp) {
//            Date attValue = ((XAttributeTimestamp) attribute).getValue();
//            HashMap<Date, Date> map;
//            if ((map = cachedTimestampAttribute.get(attName)) == null) {
//                map = new HashMap<>();
//                cachedTimestampAttribute.put(attName, map);
//            }
//            if ((attValue = map.get(attValue)) == null) {
//                attValue = ((XAttributeTimestamp) attribute).getValue();
//                map.put(attValue, attValue);
//            }
//            att = new XAttributeTimestampImpl(attName, attValue);
        } else if (attribute instanceof XAttributeContinuous) {
            Double attValue = ((XAttributeContinuous) attribute).getValue();
            HashMap<Double, Double> map;
            if ((map = cachedContinuousAttribute.get(attName)) == null) {
                map = new HashMap<Double, Double>();
                cachedContinuousAttribute.put(attName, map);
            }
            if ((attValue = map.get(attValue)) == null) {
                attValue = ((XAttributeContinuous) attribute).getValue();
                map.put(attValue, attValue);
            }
            att = factory.createAttributeContinuous(attName, attValue, null);
        } else if (attribute instanceof XAttributeDiscrete) {
            Long attValue = ((XAttributeDiscrete) attribute).getValue();
            HashMap<Long, Long> map;
            if ((map = cachedDiscreteAttribute.get(attName)) == null) {
                map = new HashMap<Long, Long>();
                cachedDiscreteAttribute.put(attName, map);
            }
            if ((attValue = map.get(attValue)) == null) {
                attValue = ((XAttributeDiscrete) attribute).getValue();
                map.put(attValue, attValue);
            }
            att = factory.createAttributeDiscrete(attName, attValue, null);
        } else {
            att = attribute;
        }

        return att;

    }

    public XEvent getXEvent(XEvent event) {
        XEvent newEvent = factory.createEvent();
        for (Map.Entry<String, XAttribute> entry : event.getAttributes().entrySet()) {
            String attName = getXAttributeName(entry.getKey());

            XAttribute att = getXAttribute(entry.getKey(), entry.getValue());

            newEvent.getAttributes().put(attName, att);
        }

        return newEvent;
    }

    public XTrace getXTrace(XTrace trace) {
        XTrace newTrace = factory.createTrace(trace.getAttributes());
        for (XEvent event : trace) {
            XEvent newEvent = getXEvent(event);
            newTrace.add(newEvent);
        }
        return newTrace;
    }

}
